@@ -16,8 +16,32 @@
 #include "input-scrub.h"
 #include "sections.h"
 
+/*
+ * The assembler can assemble the trailing +/- by setting either the Y-bit or
+ * the AT-bits.  The default is setting the Y-bit and is the same as specifying:
+ *  -static_branch_prediction_Y_bit
+ *	Treat a single trailing '+' or '-' after a conditional PowerPC branch
+ *	instruction as a static branch prediction that sets the Y-bit in the
+ *	opcode.  Pairs of trailing "++" or "--" always set the AT-bits. This is
+ *	the default for Mac OS X.
+ * This can be changed by specifying:
+ * -static_branch_prediction_AT_bits
+ *	Treat a single trailing '+' or '-' after a conditional PowerPC branch
+ *	instruction as a static branch prediction sets the AT-bits in the
+ *	opcode. Pairs of trailing "++" or "--" always set the AT-bits but with
+ *	this option a warning is issued if this syntax is used.  With this flag 
+ *	the assembler behaves like the IBM tools.
+ */
+enum static_branch_prediction {
+    STATIC_BRANCH_PREDICTION_Y_BIT,
+    STATIC_BRANCH_PREDICTION_AT_BITS
+};
+static int static_branch_prediction_specified = 0;
+static enum static_branch_prediction static_branch_prediction =
+    STATIC_BRANCH_PREDICTION_Y_BIT;
+
 /* relocation type for internal assembler use only for LIKELY_{,NOT_}TAKEN */
-#define PPC_RELOC_BR14_predicted (127)
+#define PPC_RELOC_BR14_predicted (0x10 | PPC_RELOC_BR14)
 enum branch_prediction {
     BRANCH_PREDICTION_NONE,
     BRANCH_PREDICTION_LIKELY_TAKEN,
@@ -41,8 +65,13 @@ int flag_gregs[32] = { 0 };
 /*
  * These are the default cputype and cpusubtype for the ppc architecture.
  */
+#ifdef ARCH64
+const cpu_type_t md_cputype = CPU_TYPE_POWERPC64;
+cpu_subtype_t md_cpusubtype = CPU_SUBTYPE_POWERPC_ALL;
+#else
 const cpu_type_t md_cputype = CPU_TYPE_POWERPC;
 cpu_subtype_t md_cpusubtype = CPU_SUBTYPE_POWERPC_ALL;
+#endif
 
 /* This is the byte sex for the ppc architecture */
 const enum byte_sex md_target_byte_sex = BIG_ENDIAN_BYTE_SEX;
@@ -397,7 +426,8 @@ static char *parse_num(
     unsigned long parcnt,
     long max_width_zero,
     long zero_only,
-    long signed_num);
+    long signed_num,
+    long bit_mask_with_1_bit_set);
 static char *parse_mbe(
     char *param,
     struct ppc_insn *insn,
@@ -422,7 +452,8 @@ void
 md_begin(void)
 {
     unsigned long i;
-    char *name, *retval;
+    char *name;
+    const char *retval;
 
 	/* initialize the opcode hash table */
 	op_hash = hash_new();
@@ -491,6 +522,28 @@ char ***vecP)
 		return(1);
 	    }
 	    break;
+	case 's':
+	    if(strcmp(*argP, "static_branch_prediction_Y_bit") == 0){
+		if(static_branch_prediction_specified &&
+		   static_branch_prediction != STATIC_BRANCH_PREDICTION_Y_BIT)
+		    as_bad("Can't specify both -static_branch_prediction_Y_bit"
+			    " and -static_branch_prediction_AT_bits");
+		static_branch_prediction_specified = 1;
+		static_branch_prediction = STATIC_BRANCH_PREDICTION_Y_BIT;
+		*argP = "";
+		return(1);
+	    }
+	    else if(strcmp(*argP, "static_branch_prediction_AT_bits") == 0){
+		if(static_branch_prediction_specified &&
+		   static_branch_prediction != STATIC_BRANCH_PREDICTION_AT_BITS)
+		    as_bad("Can't specify both -static_branch_prediction_Y_bit"
+			    " and -static_branch_prediction_AT_bits");
+		static_branch_prediction_specified = 1;
+		static_branch_prediction = STATIC_BRANCH_PREDICTION_AT_BITS;
+		*argP = "";
+		return(1);
+	    }
+	    break;
 	}
 	return(0);
 }
@@ -519,7 +572,7 @@ int reg)
 	SKIP_WHITESPACE();
 	if ( * input_line_pointer != ',' ) {
 		*end_name = 0;
-		as_warn("Expected comma after name \"%s\"", name);
+		as_bad("Expected comma after name \"%s\"", name);
 		*end_name = delim;
 		ignore_rest_of_line();
 		return;
@@ -580,7 +633,7 @@ int ignore)
 
 	reg = get_absolute_expression();
 	if(reg < 0 || reg >= 32)
-	    as_warn("register number (%d) out of range (0-31) for .flag_reg",
+	    as_bad("register number (%d) out of range (0-31) for .flag_reg",
 		    reg);
 	demand_empty_rest_of_line();
 	flag_registers = 1;
@@ -600,7 +653,7 @@ int ignore)
 
 	reg = get_absolute_expression();
 	if(reg < 0 || reg >= 32)
-	    as_warn("register number (%d) out of range (0-31) for .noflag_reg",
+	    as_bad("register number (%d) out of range (0-31) for .noflag_reg",
 		    reg);
 	demand_empty_rest_of_line();
 	flag_gregs[reg] = 0;
@@ -629,6 +682,7 @@ char *op)
 
     static char *file_spec;
     static unsigned long line_spec;
+    static int syntax_warning_issued_for_AT_bits = 0;
 
 	/*
 	 * Pick up the instruction and any trailing branch prediction character
@@ -641,21 +695,41 @@ char *op)
 	    end_op = param;
 	if(*end_op == '+'){
 	    if(end_op != start_op && end_op[-1] == '+'){
+		if(static_branch_prediction ==
+		   STATIC_BRANCH_PREDICTION_AT_BITS &&
+		   syntax_warning_issued_for_AT_bits == 0){
+		    as_warn("branch prediction ++/-- syntax always sets the "
+			       "AT-bits");
+		    syntax_warning_issued_for_AT_bits = 1;
+		}
 		prediction = BRANCH_PREDICTION_VERY_LIKELY_TAKEN;
 		end_op[-1] = '\0';
 	    }
 	    else{
-		prediction = BRANCH_PREDICTION_LIKELY_TAKEN;
+		if(static_branch_prediction == STATIC_BRANCH_PREDICTION_AT_BITS)
+		    prediction = BRANCH_PREDICTION_VERY_LIKELY_TAKEN;
+		else
+		    prediction = BRANCH_PREDICTION_LIKELY_TAKEN;
 		*end_op = '\0';
 	    }
 	}
 	else if(*end_op == '-'){
 	    if(end_op != start_op && end_op[-1] == '-'){
+		if(static_branch_prediction ==
+		   STATIC_BRANCH_PREDICTION_AT_BITS &&
+		   syntax_warning_issued_for_AT_bits == 0){
+		    as_warn("branch prediction ++/-- syntax always sets the "
+			       "AT-bits");
+		    syntax_warning_issued_for_AT_bits = 1;
+		}
 		prediction = BRANCH_PREDICTION_VERY_LIKELY_NOT_TAKEN;
 		end_op[-1] = '\0';
 	    }
 	    else{
-		prediction = BRANCH_PREDICTION_LIKELY_NOT_TAKEN;
+		if(static_branch_prediction == STATIC_BRANCH_PREDICTION_AT_BITS)
+		    prediction = BRANCH_PREDICTION_VERY_LIKELY_NOT_TAKEN;
+		else
+		    prediction = BRANCH_PREDICTION_LIKELY_NOT_TAKEN;
 		*end_op = '\0';
 	    }
 	}
@@ -664,7 +738,7 @@ char *op)
 
 	/* try to find the instruction in the hash table */
 	if((format = (struct ppc_opcode *)hash_find(op_hash, op)) == NULL){
-	    as_warn("Invalid mnemonic '%s'", op);
+	    as_bad("Invalid mnemonic '%s'", op);
 	    return;
 	}
 
@@ -681,13 +755,13 @@ char *op)
 	    else{
 		if(retry == 0){
 		    if(error_param_message != NULL)
-			as_warn(error_param_message, error_param_count + 1);
+			as_bad(error_param_message, error_param_count + 1);
 		    else
-			as_warn("Parameter syntax error (parameter %lu)",
+			as_bad("Parameter syntax error (parameter %lu)",
 				error_param_count + 1);
 		}
 		else
-		    as_warn("Parameter syntax error");
+		    as_bad("Parameter syntax error");
 		return;
 	    }
 	}
@@ -711,9 +785,9 @@ char *op)
 	   (insn.opcode & 0xfc000003) == 0xe8000001 || /* ldu */
 	   (insn.opcode & 0xfc0007fe) == 0x7c00006a){  /* ldux */
 	    if(RA(insn.opcode) == 0)
-		as_warn("Invalid form of the instruction (RA must not be 0)");
+		as_bad("Invalid form of the instruction (RA must not be 0)");
 	    if(RA(insn.opcode) == RT(insn.opcode))
-		as_warn("Invalid form of the instruction (RA must not the same "
+		as_bad("Invalid form of the instruction (RA must not the same "
 			"as RT)");
 	}
 	/*
@@ -741,7 +815,7 @@ char *op)
 	   (insn.opcode & 0xfc0007fe) == 0x7c0002ac || /* dst, dstt */
 	   (insn.opcode & 0xfc0007fe) == 0x7c0002ec){  /* dstst, dststt */
 	    if(RA(insn.opcode) == 0)
-		as_warn("Invalid form of the instruction (RA must not be 0)");
+		as_bad("Invalid form of the instruction (RA must not be 0)");
 	}
 	/*
 	 * For the following instruction lmw if RA is in the range of
@@ -750,7 +824,7 @@ char *op)
 	 */
 	if((insn.opcode & 0xfc000000) == 0xb8000000){ /* lmw */
 	    if(RT(insn.opcode) <= RA(insn.opcode))
-		as_warn("Invalid form of the instruction (RA is in the range "
+		as_bad("Invalid form of the instruction (RA is in the range "
 			"of registers to be loaded)");
 	}
 	/*
@@ -766,11 +840,11 @@ char *op)
 		nr = (nb + 3) / 4;
 		if(RA(insn.opcode) >= RT(insn.opcode) &&
 		   RA(insn.opcode) <= RT(insn.opcode) + nr - 1)
-		    as_warn("Invalid form of the instruction (RA is in the "
+		    as_bad("Invalid form of the instruction (RA is in the "
 			    "range of registers to be loaded)");
 		if(RT(insn.opcode) + nr - 1 > 31 &&
 		   RA(insn.opcode) < (RT(insn.opcode) + nr - 1) - 31)
-		    as_warn("Invalid form of the instruction (RA is in the "
+		    as_bad("Invalid form of the instruction (RA is in the "
 			    "range of registers to be loaded)");
 	}
 	/*
@@ -780,12 +854,13 @@ char *op)
 	 */
 	if((insn.opcode & 0xfc0007fe) == 0x7c00042a){  /* lswx */
 	    if(RT(insn.opcode) == RA(insn.opcode))
-		as_warn("Invalid form of the instruction (RT must not the same "
+		as_bad("Invalid form of the instruction (RT must not the same "
 			"as RA)");
 	    if(RT(insn.opcode) == RB(insn.opcode))
-		as_warn("Invalid form of the instruction (RT must not the same "
+		as_bad("Invalid form of the instruction (RT must not the same "
 			"as RB)");
 	}
+#if !defined(ARCH64)
 	/*
 	 * The 64-bit compares are invalid on 32-bit implementations.  Since
 	 * we don't expect to ever use the 620 all 64-bit instructions require
@@ -796,10 +871,12 @@ char *op)
 	    (insn.opcode & 0xfc400000) == 0x2c000000 ||  /* cmpi */
 	    (insn.opcode & 0xfc400000) == 0x28000000) && /* cmpli */
 	   (insn.opcode & 0x00200000) == 0x00200000 &&   /* the L bit */
-	   !force_cpusubtype_ALL){
-	    as_warn("Invalid form of the instruction (64-bit compares not "
+	   !force_cpusubtype_ALL &&
+	   archflag_cpusubtype != CPU_SUBTYPE_POWERPC_970){
+	    as_bad("Invalid form of the instruction (64-bit compares not "
 		    "allowed without -force_cpusubtype_ALL option)");
 	}
+#endif /* !defined(ARCH64) */
 	/*
 	 * For branch conditional instructions certian BO fields are reserved.
 	 * These are flagged as invalid forms unless the -force_cpusubtype_ALL
@@ -816,7 +893,7 @@ char *op)
 	     */
 	    if((insn.opcode & 0x02800000) == 0x02800000 && /* 1z1zz */
 	       (insn.opcode & 0x01600000) != 0x00000000){
-		as_warn("Invalid form of the instruction (reserved bits in the "
+		as_bad("Invalid form of the instruction (reserved bits in the "
 			"BO field must be zero without -force_cpusubtype_ALL "
 			"option)");
 	    }
@@ -838,7 +915,7 @@ char *op)
 		    val = (insn.opcode & (0x1f << format->ops[i].offset)) >>
 			  format->ops[i].offset;
 		    if(flag_gregs[val])
-			as_warning("flagged register r%lu used", val);
+			as_bad("flagged register r%lu used", val);
 		}
 	    }
 	}
@@ -848,28 +925,39 @@ char *op)
 	 */
 	if(format->cpus != 0 && !force_cpusubtype_ALL){
 	    if(no_ppc601 == 1 && format->cpus == CPU601)
-		as_warning("not allowed 601 instruction \"%s\"", format->name);
-	    if((format->cpus & IMPL64) == IMPL64){
-		as_bad("instruction is only for 64-bit implementations (not "
-		       "allowed without -force_cpusubtype_ALL option)");
+		as_bad("not allowed 601 instruction \"%s\"", format->name);
+#if !defined(ARCH64)
+	    if((format->cpus & IMPL64) == IMPL64
+		&& archflag_cpusubtype != CPU_SUBTYPE_POWERPC_970
+		){
+		as_bad("%s instruction is only for 64-bit implementations (not "
+		       "allowed without -force_cpusubtype_ALL option)",
+		       format->name);
 	    }
-	    if(format->cpus == OPTIONAL){
-		as_bad("instruction is optional for the PowerPC (not "
-		       "allowed without -force_cpusubtype_ALL option)");
+	    if((format->cpus & OPTIONAL) == OPTIONAL){
+		if((format->cpus & CPU970) == CPU970 &&
+		   archflag_cpusubtype != CPU_SUBTYPE_POWERPC_970)
+		    as_bad("%s instruction is optional for the PowerPC (not "
+			   "allowed without -force_cpusubtype_ALL option)",
+			   format->name);
 	    }
 	    if(format->cpus == VMX &&
 	       (archflag_cpusubtype != CPU_SUBTYPE_POWERPC_7400 &&
-	        archflag_cpusubtype != CPU_SUBTYPE_POWERPC_7450)){
-		as_bad("vector instruction is optional for the PowerPC (not "
-		       "allowed without -force_cpusubtype_ALL option)");
+	        archflag_cpusubtype != CPU_SUBTYPE_POWERPC_7450 &&
+	        archflag_cpusubtype != CPU_SUBTYPE_POWERPC_970)){
+		as_bad("%s vector instruction is optional for the PowerPC (not "
+		       "allowed without -force_cpusubtype_ALL option)",
+		       format->name);
 	    }
-	    else if(md_cpusubtype == CPU_SUBTYPE_POWERPC_ALL){
+	    else
+#endif /* !defined(ARCH64) */
+	    if(md_cpusubtype == CPU_SUBTYPE_POWERPC_ALL){
 		switch(format->cpus){
 		case CPU601:
 		    if(archflag_cpusubtype != -1 &&
 		       archflag_cpusubtype != CPU_SUBTYPE_POWERPC_601)
-			as_bad("601 instruction not allowed with -arch %s",
-			       specific_archflag);
+			as_bad("%s 601 instruction not allowed with -arch %s",
+			       format->name, specific_archflag);
 		    else{
 			file_spec = logical_input_file ?
 				    logical_input_file : physical_input_file;
@@ -885,8 +973,8 @@ char *op)
 		case CPU601:
 		    if(archflag_cpusubtype != -1 &&
 		       archflag_cpusubtype != CPU_SUBTYPE_POWERPC_601)
-			as_bad("601 instruction not allowed with -arch %s",
-			       specific_archflag);
+			as_bad("%s 601 instruction not allowed with -arch %s",
+			       format->name, specific_archflag);
 		    else{
 			if(md_cpusubtype != CPU_SUBTYPE_POWERPC_601)
 			    as_bad("more than one implementation specific "
@@ -965,7 +1053,7 @@ char *op)
 		    insn.reloc);
 	    break;
 	default:
-	    as_warn("Unknown relocation type");
+	    as_bad("Unknown relocation type");
 	    break;
 	}
 	if(insn.jbsr_exp.X_add_symbol != NULL){
@@ -1043,19 +1131,22 @@ enum branch_prediction prediction)
 		param = parse_crf(param, insn, format, parcnt);
 		break;
 	    case SNUM:
-		param = parse_num(param, insn, format, parcnt, 0, 0, 1);
+		param = parse_num(param, insn, format, parcnt, 0, 0, 1, 0);
+		break;
+	    case FXM:
+		param = parse_num(param, insn, format, parcnt, 0, 0, 0, 1);
 		break;
 	    case NUM:
-		param = parse_num(param, insn, format, parcnt, 0, 0, 0);
+		param = parse_num(param, insn, format, parcnt, 0, 0, 0, 0);
 		break;
 	    case NUM0:
-		param = parse_num(param, insn, format, parcnt, 1, 0, 0);
+		param = parse_num(param, insn, format, parcnt, 1, 0, 0, 0);
 		break;
 	    case MBE:
 		param = parse_mbe(param, insn, format, parcnt);
 		break;
 	    case ZERO:
-		param = parse_num(param, insn, format, parcnt, 0, 1, 0);
+		param = parse_num(param, insn, format, parcnt, 0, 1, 0, 0);
 		break;
 	    case sh:
 		param = parse_sh(param, insn, format, parcnt);
@@ -1087,7 +1178,8 @@ enum branch_prediction prediction)
 		     * in md_number_to_imm() if the reloc is
 		     * PPC_RELOC_BR14_predicted.
 		     */
-		    insn->reloc = PPC_RELOC_BR14_predicted;
+		    if(insn->reloc == PPC_RELOC_BR14)
+			insn->reloc = PPC_RELOC_BR14_predicted;
 		    if(prediction == BRANCH_PREDICTION_LIKELY_TAKEN)
 			insn->opcode |= Y_BIT;
 		    else{ /* prediction == BRANCH_PREDICTION_LIKELY_NOT_TAKEN */
@@ -1111,7 +1203,7 @@ enum branch_prediction prediction)
 			else
 			    insn->opcode |= 0x00400000; /* AT == 10 */
 		    }
-		    else if(bo == 0x04 || bo == 0x14){
+		    else if(bo == 0x10 || bo == 0x12){
 			/*
 			 * For 'decrement the CTR, then branch if the
 			 * decremented CTR is non-zero or zero' the AT bits are
@@ -1151,7 +1243,7 @@ struct ppc_insn *insn,
 struct ppc_opcode *format,
 int parcnt)
 {
-    unsigned long val;
+    signed_target_addr_t val;
     char *end, *saveptr, *saveparam;
     segT seg;
 
@@ -1509,7 +1601,7 @@ unsigned long parcnt)
 
 	    while(isdigit(*param))
 		if((val = val * 10 + *param++ - '0') >=
-		   1 << format->ops[parcnt].width)
+		   (unsigned long)(1 << format->ops[parcnt].width))
 		return(NULL);
 
 	    if(format->ops[parcnt].type == G0REG && val == 0){
@@ -1554,7 +1646,7 @@ struct ppc_insn *insn,
 struct ppc_opcode *format,
 unsigned long parcnt)
 {
-    int val;
+    signed_target_addr_t val;
     unsigned long i;
     char *saveptr, save_c;
     expressionS exp;
@@ -1624,7 +1716,7 @@ struct ppc_insn *insn,
 struct ppc_opcode *format,
 unsigned long parcnt)
 {
-    int val;
+    signed_target_addr_t val;
     unsigned long i, j;
     char *saveptr, save_c, *plus, save_plus;
     expressionS exp;
@@ -1719,7 +1811,7 @@ struct ppc_insn *insn,
 struct ppc_opcode *format,
 unsigned long parcnt)
 {
-    int val;
+    signed_target_addr_t val;
     unsigned long i;
     char *saveptr, save_c;
     expressionS exp;
@@ -1794,9 +1886,11 @@ struct ppc_opcode *format,
 unsigned long parcnt,
 long max_width_zero,
 long zero_only,
-long signed_num)
+long signed_num,
+long bit_mask_with_1_bit_set)
 {
-    int val, max, min;
+    signed_target_addr_t val;
+    int i, max, min, mask, temp;
     char *saveptr, save_c;
     expressionS exp;
     segT seg;
@@ -1826,15 +1920,49 @@ long signed_num)
 	if(signed_num){
 	    max = (1 << (format->ops[parcnt].width - 1)) - 1;
 	    min = (0xffffffff << (format->ops[parcnt].width - 1));
+	    temp = val;
+	    if(temp > max || temp < min){
+		error_param_message = "Parameter error: expression out "
+				      "of range (parameter %lu)";
+		return(NULL);
+	    }
 	}
 	else{
 	    max = (1 << (format->ops[parcnt].width)) - 1;
-	    min = 0;
+	    if(val > max){
+		error_param_message = "Parameter error: expression out "
+				      "of range (parameter %lu)";
+		return(NULL);
+	    }
 	}
-	if(val > max || val < min){
-	    error_param_message = "Parameter error: expression out "
-				  "of range (parameter %lu)";
-	    return(NULL);
+	if(bit_mask_with_1_bit_set){
+	    mask = 1;
+	    for(i = 0; i < format->ops[parcnt].width; i++){
+		if(mask & val)
+		    break;
+		mask = mask << 1;
+	    }
+	    /*
+	     * If this is the mtcrf opcode (0x7c000120) and val is not zero and
+	     * has exactly one bit set then use the new form of the mtcrf
+	     * opcode.  This has bit 0x00100000 set and the FXM field is a bit
+	     * mask. Else use the old form without bit 0x00100000 set.
+	     */ 
+	    if(insn->opcode == 0x7c000120){
+		if(val != 0 && val == mask)
+		    insn->opcode |= 0x00100000;
+	    }
+	    else{
+		/*
+		 * For instructions other than mtcrf if exactly one bit in val
+		 * is not set it is an error.
+		 */
+		if(val == 0 || val != mask){
+		    error_param_message = "Parameter error: expression must "
+				  "have exactly one bit set (parameter %lu)";
+		    return(NULL);
+		}
+	    }
 	}
 	if(zero_only == 1 && val != 0){
 	    error_param_message = "Parameter error: expression must have a "
@@ -1870,7 +1998,7 @@ struct ppc_insn *insn,
 struct ppc_opcode *format,
 unsigned long parcnt)
 {
-    int val;
+    signed_target_addr_t val;
     char *saveptr, save_c;
     expressionS exp;
     segT seg;
@@ -1968,7 +2096,7 @@ struct ppc_insn *insn,
 struct ppc_opcode *format,
 unsigned long parcnt)
 {
-    int val;
+    signed_target_addr_t val;
     char *saveptr, save_c;
     expressionS exp;
     segT seg;
@@ -2029,7 +2157,7 @@ struct ppc_insn *insn,
 struct ppc_opcode *format,
 unsigned long parcnt)
 {
-    int val;
+    signed_target_addr_t val;
     char *saveptr, save_c;
     expressionS exp;
     segT seg;
@@ -2082,17 +2210,22 @@ unsigned long parcnt)
 
 /*
  * md_number_to_chars() is the target machine dependent routine that puts out
- * a binary value of size 4, 2, or 1 bytes into the specified buffer.  This is
- * done in the target machine's byte sex.  In this case the byte order is
+ * a binary value of size 8, 4, 2, or 1 bytes into the specified buffer.  This
+ * is done in the target machine's byte sex.  In this case the byte order is
  * big endian.
  */
 void
 md_number_to_chars(
 char *buf,
-long val,
+signed_expr_t val,
 int nbytes)
 {
 	switch(nbytes){
+	case 8:
+	    *buf++ = val >> 56;
+	    *buf++ = val >> 48;
+	    *buf++ = val >> 40;
+	    *buf++ = val >> 32;
 	case 4:
 	    *buf++ = val >> 24;
 	    *buf++ = val >> 16;
@@ -2118,7 +2251,7 @@ int nbytes)
 void
 md_number_to_imm(
 unsigned char *buf,
-long val,
+signed_expr_t val,
 int nbytes,
 fixS *fixP,
 int nsect)
@@ -2128,6 +2261,11 @@ int nsect)
 	if(fixP->fx_r_type == NO_RELOC ||
 	   fixP->fx_r_type == PPC_RELOC_VANILLA){
 	    switch(nbytes){
+            case 8:
+                *buf++ = val >> 56;
+                *buf++ = val >> 48;
+                *buf++ = val >> 40;
+                *buf++ = val >> 32;
 	    case 4:
 		*buf++ = val >> 24;
 		*buf++ = val >> 16;
@@ -2161,18 +2299,24 @@ int nsect)
 
 	case PPC_RELOC_LO14:
 	    buf[2] = val >> 8;
-	    buf[3] = val & 0xfc;
+	    buf[3] |= val & 0xfc;
 	    break;
 
 	case PPC_RELOC_BR14:
 	case PPC_RELOC_BR14_predicted:
 	    if(fixP->fx_pcrel)
 		val += 4;
-	    if((val & 0xffff8000) && ((val & 0xffff8000) != 0xffff8000))
-		as_warn("Fixup of %ld too large for field width of 16 bits",
-                        val);
-	    if((val & 0x3) != 0)
-		as_warn("Fixup of %ld is not to a 4 byte address", val);
+	    if((val & 0xffff8000) && ((val & 0xffff8000) != 0xffff8000)){
+		layout_file = fixP->file;
+		layout_line = fixP->line;
+		as_bad("Fixup of %lld too large for field width of 16 "
+			"bits", val);
+	    }
+	    if((val & 0x3) != 0){
+		layout_file = fixP->file;
+		layout_line = fixP->line;
+		as_bad("Fixup of %lld is not to a 4 byte address", val);
+	    }
 	    /*
 	     * Note PPC_RELOC_BR14 are only used with bc, "branch conditional"
 	     * instructions.  The Y_BIT was previously set assuming the
@@ -2202,19 +2346,22 @@ int nsect)
 	    }
 	    buf[2] = val >> 8;
 	    buf[3] |= val & 0xfc;
-	    /* change any PPC_RELOC_BR14_predicted back to PPC_RELOC_BR14
-	       before it is used to create a relocation entry. */
-	    fixP->fx_r_type = PPC_RELOC_BR14;
 	    break;
 
 	case PPC_RELOC_BR24:
 	    if(fixP->fx_pcrel)
 		val += 4;
-	    if((val & 0xfc000000) && ((val & 0xfc000000) != 0xfc000000))
-		as_warn("Fixup of %ld too large for field width of 26 bits",
-                        val);
-	    if((val & 0x3) != 0)
-		as_warn("Fixup of %ld is not to a 4 byte address", val);
+	    if((val & 0xfc000000) && ((val & 0xfc000000) != 0xfc000000)){
+		layout_file = fixP->file;
+		layout_line = fixP->line;
+		as_bad("Fixup of %lld too large for field width of 26 "
+			"bits", val);
+	    }
+	    if((val & 0x3) != 0){
+		layout_file = fixP->file;
+		layout_line = fixP->line;
+		as_bad("Fixup of %lld is not to a 4 byte address", val);
+	    }
 	    buf[0] |= (val >> 24) & 0x03;
 	    buf[1] = val >> 16;
 	    buf[2] = val >> 8;
@@ -2226,7 +2373,9 @@ int nsect)
 	    break;
 
 	default:
-	    as_warn("Bad relocation type");
+	    layout_file = fixP->file;
+	    layout_line = fixP->line;
+	    as_bad("Bad relocation type");
 	    break;
 	}
 }
@@ -2286,7 +2435,7 @@ md_estimate_size_before_relax(
 fragS *fragP,
 int segment_type)
 {
-	as_warn("Relaxation should never occur");
+	as_bad("Relaxation should never occur");
 	return(sizeof(long));
 }
 
@@ -2296,5 +2445,5 @@ void
 md_convert_frag(
 fragS *fragP)
 {
-	as_warn("Relaxation should never occur");
+	as_bad("Relaxation should never occur");
 }
